import math
from collections import defaultdict

import torch
from torch.optim.optimizer import Optimizer, required


class RAdam(Optimizer):
    """RAdam optimizer, a theoretically sound variant of Adam.

    Source: `LiyuanLucasLiu/RAdam <https://github.com/LiyuanLucasLiu/RAdam/blob/master/radam/radam.py>`_

    Under Apache License 2.0
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, degenerated_to_sgd=True):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(
                "Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(
                "Invalid beta parameter at index 1: {}".format(betas[1]))

        self.degenerated_to_sgd = degenerated_to_sgd
        if isinstance(params, (list, tuple)) and len(params) > 0 and isinstance(params[0], dict):
            for param in params:
                if 'betas' in param and (param['betas'][0] != betas[0] or param['betas'][1] != betas[1]):
                    param['buffer'] = [[None, None, None] for _ in range(10)]
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay,
                        buffer=[[None, None, None] for _ in range(10)])
        super(RAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(RAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError(
                        'RAdam does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(
                        p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                state['step'] += 1
                buffered = group['buffer'][int(state['step'] % 10)]
                if state['step'] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2 ** state['step']
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * \
                        state['step'] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma

                    # more conservative since it's an approximated value
                    if N_sma >= 5:
                        step_size = math.sqrt((1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) /
                                              N_sma * N_sma_max / (N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    elif self.degenerated_to_sgd:
                        step_size = 1.0 / (1 - beta1 ** state['step'])
                    else:
                        step_size = -1
                    buffered[2] = step_size

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay']
                                         * group['lr'], p_data_fp32)
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    p_data_fp32.addcdiv_(
                        exp_avg, denom, value=-step_size * group['lr'])
                    p.data.copy_(p_data_fp32)
                elif step_size > 0:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay']
                                         * group['lr'], p_data_fp32)
                    p_data_fp32.add_(exp_avg, alpha=-step_size * group['lr'])
                    p.data.copy_(p_data_fp32)

        return loss


class Lookahead(Optimizer):
    '''Lookahead Wrapper

    * Code: `lonePatient/lookahead_pytorch <https://github.com/lonePatient/lookahead_pytorch/blob/master/optimizer.py>`_
    * Paper: `Lookahead Optimizer <https://arxiv.org/abs/1907.08610>`_

    Works best with `LookaheadCallback` or `LookaheadModelCheckpoint`.

    Args:
        optimizer (Optimizer): The inner optimizer.
        alpha (float, optional): The linear interpolation factor. 1.0 recovers the inner optimizer. Defaults to 0.5.
        k (int, optional): The number of lookahead steps. Defaults to 6.
        pullback_momentum (str, optional): Change to inner optimizer momentum on interpolation update. Defaults to "none".

    .. note::

       Currently `pullback_momentum` only supports SGD optimizers with momentum.

    Raises:
        ValueError: Invalid slow update rate or invalid lookahead steps

    Example:

        >>> model = torch.nn.Linear(10, 1)
        >>> optimizer = Lookahead(
        ...     torch.optim.SGD(model.parameters(), momentum=0.9, lr=0.1),
        ...     alpha=0.5, k=6, pullback_momentum="pullback")
        ...
        >>> for _ in range(10):
        ...     optimizer.zero_grad()
        ...     loss = model(torch.rand(10))
        ...     loss.backward()
        ...     optimizer.step()
        ...
    '''

    def __init__(self, optimizer: Optimizer, alpha: float = 0.5, k: int = 6, pullback_momentum: str = "none"):
        if not 0.0 <= alpha <= 1.0:
            raise ValueError(f'Invalid slow update rate: {alpha}')
        if not 1 <= k:
            raise ValueError(f'Invalid lookahead steps: {k}')
        self.optimizer = optimizer
        self.param_groups = self.optimizer.param_groups
        self.alpha = alpha
        self.k = k
        self.step_counter = 0
        assert pullback_momentum in ["reset", "pullback", "none"]
        self.pullback_momentum = pullback_momentum
        self.state = defaultdict(dict)
        self.defaults = {}

        # Cache the current optimizer parameters
        for group in self.optimizer.param_groups:
            for p in group['params']:
                param_state = self.state[p]
                param_state['cached_params'] = torch.zeros_like(p.data)
                param_state['cached_params'].copy_(p.data)

    def __getstate__(self):
        return {
            'state': self.state,
            'optimizer': self.optimizer,
            'alpha': self.alpha,
            'step_counter': self.step_counter,
            'k': self.k,
            'pullback_momentum': self.pullback_momentum
        }

    def zero_grad(self):
        self.optimizer.zero_grad()

    def state_dict(self):
        return self.optimizer.state_dict()

    def load_state_dict(self, state_dict):
        self.optimizer.load_state_dict(state_dict)

    def _backup_and_load_cache(self):
        """Useful for performing evaluation on the slow weights (which typically generalize better)
        """
        for group in self.optimizer.param_groups:
            for p in group['params']:
                param_state = self.state[p]
                param_state['backup_params'] = torch.zeros_like(p.data)
                param_state['backup_params'].copy_(p.data)
                p.data.copy_(param_state['cached_params'])

    def _clear_and_load_backup(self):
        for group in self.optimizer.param_groups:
            for p in group['params']:
                param_state = self.state[p]
                p.data.copy_(param_state['backup_params'])
                del param_state['backup_params']

    def step(self, closure=None):
        """Performs a single Lookahead optimization step.
        """
        loss = self.optimizer.step(closure)
        self.step_counter += 1

        if self.step_counter >= self.k:
            self.step_counter = 0
            # Lookahead and cache the current optimizer parameters
            for group in self.optimizer.param_groups:
                for p in group['params']:
                    param_state = self.state[p]
                    p.data.mul_(self.alpha).add_(1.0 - self.alpha,
                                                 param_state['cached_params'])  # crucial line
                    param_state['cached_params'].copy_(p.data)
                    if self.pullback_momentum == "pullback":
                        if "cached_mom" in param_state:
                            internal_momentum = self.optimizer.state[p]["momentum_buffer"]
                            self.optimizer.state[p]["momentum_buffer"] = internal_momentum.mul_(self.alpha).add_(
                                1.0 - self.alpha, param_state["cached_mom"])
                        param_state["cached_mom"] = self.optimizer.state[p]["momentum_buffer"]
                    elif self.pullback_momentum == "reset":
                        self.optimizer.state[p]["momentum_buffer"] = torch.zeros_like(
                            p.data)

        return loss



import warnings
from typing import Optional, Any

import torch
import numpy as np
from scipy.stats import spearmanr
from pytorch_lightning.metrics import Metric
from pytorch_lightning.utilities import rank_zero_warn
from sklearn.metrics import fbeta_score, roc_auc_score
from sklearn.exceptions import UndefinedMetricWarning


class GlobalMetric(Metric):
    def __init__(
        self,
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: Optional[Any] = None,
    ):
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
        )
        self.add_state("preds", default=[], dist_reduce_fx=None)
        self.add_state("target", default=[], dist_reduce_fx=None)

        rank_zero_warn(
            "This Metric will save all targets and predictions in buffer."
            " For large datasets this may lead to large memory footprint."
        )

    def update(self, preds: torch.Tensor, target: torch.Tensor):
        """
        Update state with predictions and targets.

        Args:
            preds: Predictions from model
            target: Ground truth values
        """
        self.preds.append(preds)
        self.target.append(target)


class AUC(GlobalMetric):
    """AUC: Area Under the ROC Curve


    **Binary mode**

    >>> auc = AUC()
    >>> auc(torch.tensor([0.3, 0.8, 0.2]), torch.tensor([0, 1, 0]))
    >>> auc(torch.tensor([0.3, 0.3, 0.9]), torch.tensor([1, 1, 0]))
    >>> round(auc.compute().item(), 2)
    0.56

    **Multi-class mode**

    This will use the first column as the negative case, and the rest collectively as the positive case.

    >>> auc = AUC()
    >>> auc(torch.tensor([[0.3, 0.8, 0.2], [0.2, 0.1, 0.1], [0.5, 0.1, 0.7]]).t(), torch.tensor([0, 1, 0]))
    >>> auc(torch.tensor([[0.3, 0.3, 0.8], [0.2, 0.6, 0.1], [0.5, 0.1, 0.1]]).t(), torch.tensor([1, 1, 0]))
    >>> round(auc.compute().item(), 2)
    0.39
    """

    def compute(self):
        target = torch.cat(self.target, dim=0).cpu().long().numpy()
        preds = torch.cat(self.preds, dim=0).float().cpu().numpy()
        if len(preds.shape) > 1:
            preds = 1 - preds[:, 0]
            target = (target != 0).astype(int)
        if len(np.unique(target)) == 1:
            return torch.tensor(0, device=self.preds[0].device)
        return torch.tensor(roc_auc_score(target, preds), device=self.preds[0].device)


class SpearmanCorrelation(GlobalMetric):
    def __init__(
        self,
        sigmoid: bool = False,
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: Optional[Any] = None,
    ):
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
        )
        self.sigmoid = sigmoid

    def compute(self):
        preds = torch.cat(self.preds, dim=0).float()
        if self.sigmoid:
            preds = torch.sigmoid(preds)
        preds = preds.cpu().numpy()
        target = torch.cat(self.target, dim=0).cpu().float().numpy()
        spearman_score = spearmanr(target, preds).correlation
        if len(np.unique(target)) == 1:
            return torch.tensor(0, device=self.preds[0].device)
        return torch.tensor(spearman_score, device=self.preds[0].device)


class FBeta(GlobalMetric):
    """The F-beta score is the weighted harmonic mean of precision and recall

    **Binary mode**

    >>> fbeta = FBeta()
    >>> fbeta(torch.tensor([0.3, 0.8, 0.2]), torch.tensor([0, 1, 0]))
    >>> fbeta(torch.tensor([0.3, 0.3, 0.9]), torch.tensor([1, 1, 0]))
    >>> round(fbeta.compute().item(), 2)
    0.88

    **Multi-class mode**

    This will use the first column as the negative case, and the rest collectively as the positive case.

    >>> fbeta = FBeta()
    >>> fbeta(torch.tensor([[0.8, 0.3, 0.7], [0.1, 0.1, 0.1], [0.1, 0.6, 0.2]]).t(), torch.tensor([0, 1, 0]))
    >>> fbeta(torch.tensor([[0.3, 0.7, 0.8], [0.2, 0.2, 0.1], [0.5, 0.1, 0.1]]).t(), torch.tensor([1, 1, 0]))
    >>> round(fbeta.compute().item(), 4)
    0.9375
    """

    def __init__(
        self,
        step: float = 0.02,
        beta: int = 2,
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: Optional[Any] = None,
    ):
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
        )
        self.step = step
        self.beta = beta

    def find_best_fbeta_threshold(self, truth, probs):
        best, best_thres = 0, -1
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UndefinedMetricWarning)
            for thres in np.arange(self.step, 1, self.step):
                current = fbeta_score(
                    truth,
                    (probs >= thres).astype("int8"),
                    beta=self.beta,
                    average="binary",
                )
                if current > best:
                    best = current
                    best_thres = thres
        return best, best_thres

    def compute(self):
        preds = torch.cat(self.preds, dim=0).float().cpu().numpy()
        target = torch.cat(self.target, dim=0).cpu().long().numpy()
        if len(preds.shape) > 1:
            preds = 1 - preds[:, 0]
            target = (target != 0).astype(int)
        if len(np.unique(target)) == 1:
            return torch.tensor(0, device=self.preds[0].device)
        best_fbeta, best_thres = self.find_best_fbeta_threshold(target, preds)
        return torch.tensor(best_fbeta, device=self.preds[0].device)



from typing import Callable, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


def _linear_combination(x, y, epsilon):
    return epsilon * x + (1 - epsilon) * y


class LabelSmoothCrossEntropy(nn.Module):
    """Cross Entropy with Label Smoothing

    Reference: `wangleiofficial/lable-smoothing-pytorch <https://github.com/wangleiofficial/label-smoothing-pytorch>`_

    The ground truth label will have a value of `1-eps` in the target vector.

    Args:
            eps (float): the smoothing factor.
    """

    def __init__(self, eps: float):
        super().__init__()
        self.eps = eps

    def forward(self, preds, targets, weight=None):
        n = preds.size()[-1]
        log_preds = F.log_softmax(preds, dim=-1)
        if weight is None:
            loss = -log_preds.sum(dim=-1).mean()
        else:
            loss = -(log_preds.sum(dim=-1) * weight.unsqueeze(0)).mean()
        nll = F.nll_loss(log_preds, targets, weight=weight)
        return _linear_combination(loss / n, nll, self.eps)


class MixupSoftmaxLoss(nn.Module):
    """A softmax loss that supports MixUp augmentation.

    It requires the input batch to be manipulated into certain format.
    Works best with MixUpCallback, CutMixCallback, and SnapMixCallback.

    Reference: `Fast.ai's implementation <https://github.com/fastai/fastai/blob/master/fastai/callbacks/mixup.py#L6>`_

    Args:
        class_weights (torch.Tensor, optional): The weight of each class. Defaults to the same weight.
        reduction (str, optional): Loss reduction method. Defaults to 'mean'.
        label_smooth_eps (float, optional): If larger than zero, use `LabelSmoothedCrossEntropy` instead of `CrossEntropy`. Defaults to 0.
    """

    def __init__(self, class_weights: Optional[torch.Tensor] = None, reduction: str = 'mean', label_smooth_eps: float = 0):
        super().__init__()
        # setattr(self.crit, 'reduction', 'none')
        self.reduction = reduction
        self.weight = class_weights
        # self.label_smooth_eps = label_smooth_eps
        if label_smooth_eps:
            self.loss_fn: Callable = LabelSmoothCrossEntropy(eps=label_smooth_eps)
        else:
            self.loss_fn = F.cross_entropy

    def forward(self, output: torch.Tensor, target):
        """The feed-forward.

        The target tensor should have three columns:

        1. the first class.
        2. the second class.
        3. the lambda value to mix the above two classes.

        Args:
            output (torch.Tensor): the model output.
            target (torch.Tensor): Shaped (batch_size, 3).

        Returns:
            torch.Tensor: the result loss
        """
        weight = self.weight
        if weight is not None:
            weight = self.weight.to(output.device)
        if len(target.size()) == 2:
            loss1 = self.loss_fn(output, target[:, 0].long(), weight=weight)
            loss2 = self.loss_fn(output, target[:, 1].long(), weight=weight)
            assert target.size(1) in (3, 4)
            if target.size(1) == 3:
                lambda_ = target[:, 2]
                d = (loss1 * lambda_ + loss2 * (1-lambda_)).mean()
            else:
                lamb_1, lamb_2 = target[:, 2], target[:, 3]
                d = (loss1 * lamb_1 + loss2 * lamb_2).mean()
        else:
            # This handles the cases without MixUp for backward compatibility
            d = self.loss_fn(output, target, weight=weight)
        if self.reduction == 'mean':
            return d.mean()
        elif self.reduction == 'sum':
            return d.sum()
        return d



import math
import warnings
from typing import Sequence, Union, Iterable, Optional, List

import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
import numpy as np

Layer = Union[torch.nn.Module, torch.nn.ModuleList]


class EMATracker:
    """Keeps the exponential moving average for a single series.

    Args:
        alpha (float, optional): the weight of the new value, by default 0.05

    Examples:

        >>> tracker = EMATracker(0.1)
        >>> tracker.update(1.)
        >>> tracker.value
        1.0
        >>> tracker.update(2.)
        >>> tracker.value # 1 * 0.9 + 2 * 0.1
        1.1
        >>> tracker.update(float('nan')) # this won't have any effect
        >>> tracker.value
        1.1
    """

    def __init__(self, alpha: float = 0.05):
        super().__init__()
        self.alpha = alpha
        self._value: Optional[Union[float, torch.Tensor]] = None

    def update(self, new_value: Union[float, torch.Tensor]):
        """Adds a new value to the tracker.

        It will ignore NaNs and raise a warning in those cases.

        Args:
            new_value (Union[float, torch.Tensor]): the incoming value.
        """
        if isinstance(new_value, torch.Tensor):
            new_value = new_value.detach()
            if torch.isnan(new_value):
                warnings.warn("NaN encountered as training loss!")
                return
        elif math.isnan(new_value):
            warnings.warn("NaN encountered as training loss!")
            return
        if self._value is None:
            self._value = new_value
        else:
            self._value = new_value * self.alpha + self._value * (1 - self.alpha)

    @property
    def value(self):
        """The smoothed value."""
        return self._value


def count_parameters(parameters: Iterable[Union[torch.Tensor, Parameter]]):
    """Count the number of parameters

    Args:
        parameters (Iterable[Union[torch.Tensor, Parameter]]): parameters you want to count.

    Returns:
        int: the number of parameters counted.

    Example:

        >>> count_parameters([torch.rand(100), torch.rand(10)])
        110
        >>> count_parameters([torch.rand(100, 2), torch.rand(10, 3)])
        230
    """
    return int(np.sum(list(p.numel() for p in parameters)))


# -----------------------------------
# Layer freezing from fast.ai v1
# -----------------------------------


def _children(m):
    return m if isinstance(m, (list, tuple)) else list(m.children())


def _set_trainable_attr(m, b):
    if isinstance(m, torch.nn.Parameter):
        m.requires_grad = b
        return
    m.trainable = b
    for p in m.parameters():
        p.requires_grad = b


def _apply_leaf(m, f):
    if isinstance(m, (torch.nn.Module, torch.nn.Parameter)):
        f(m)
        if isinstance(m, torch.nn.Parameter):
            return
    c = _children(m)
    if len(c) > 0:
        for l in c:
            _apply_leaf(l, f)


def set_trainable(layer: Layer, trainable: bool):
    """Freeze or unfreeze all parameters in the layer.

    Args:
        layer (Union[torch.nn.Module, torch.nn.ModuleList]): the target layer
        trainable (bool): True to unfreeze; False to freeze

    Example:

        >>> model = nn.Sequential(nn.Linear(10, 100), nn.Linear(100, 1))
        >>> model[0].weight.requires_grad
        True
        >>> set_trainable(model, False)
        >>> model[0].weight.requires_grad
        False
        >>> set_trainable(model, True)
        >>> model[0].weight.requires_grad
        True
    """
    _apply_leaf(layer, lambda m: _set_trainable_attr(m, trainable))


def freeze_layers(layer_groups: Sequence[Layer], freeze_flags: Sequence[bool]):
    """Freeze or unfreeze groups of layers

    Args:
        layer_groups (Sequence[Layer]): the target lists of layers
        freeze_flags (Sequence[bool]): the corresponding trainable flags


    .. warning::

        The value in `freeze_flag` has the opposite meaning as in `trainable` of `set_trainable`.

        Set True to freeze; False to unfreeze.

    Examples:

        >>> model = nn.Sequential(nn.Linear(10, 100), nn.Linear(100, 1))
        >>> freeze_layers([model[0], model[1]], [True, False])
        >>> model[0].weight.requires_grad
        False
        >>> model[1].weight.requires_grad
        True
        >>> freeze_layers([model[0], model[1]], [False, True])
        >>> model[0].weight.requires_grad
        True
        >>> model[1].weight.requires_grad
        False
    """
    assert len(freeze_flags) == len(layer_groups)
    for layer, flag in zip(layer_groups, freeze_flags):
        set_trainable(layer, not flag)


# -----------------------------------------------------------
# Separate BatchNorm2d and GroupNorm paremeters from others
# -----------------------------------------------------------


def separate_parameters(
    module: Union[Parameter, nn.Module, List[nn.Module]],
    skip_list: Sequence[str] = ("bias",),
):
    """Separate BatchNorm2d, GroupNorm, and LayerNorm paremeters from others

    Args:
        module (Union[Parameter, nn.Module, List[nn.Module]]): to be separated.

    Returns:
        Tuple[List[Parameter], List[Parameter]]: lists of decay and no-decay parameters.

    Example:

        >>> model = nn.Sequential(nn.Linear(100, 10, bias=True), nn.BatchNorm1d(10))
        >>> _ = nn.init.constant_(model[0].weight, 2.)
        >>> _ = nn.init.constant_(model[0].bias, 1.)
        >>> _ = nn.init.constant_(model[1].weight, 1.)
        >>> _ = nn.init.constant_(model[1].bias, 1.)
        >>> model[0].weight.data.sum().item()
        2000.0
        >>> model[0].bias.data.sum().item()
        10.0
        >>> model[1].weight.data.sum().item()
        10.0
        >>> model[1].bias.data.sum().item()
        10.0
        >>> decay, no_decay = separate_parameters(model) # separate the parameters
        >>> np.sum([x.sum().detach().numpy() for x in decay]) # nn.Linear
        2000.0
        >>> np.sum([x.sum().detach().numpy() for x in no_decay]) # nn.BatchNorm1d
        30.0
        >>> optimizer = torch.optim.AdamW([{
        ...    "params": decay, "weight_decay": 0.1
        ... }, {
        ...    "params": no_decay, "weight_decay": 0
        ... }], lr = 1e-3)
    """
    decay, no_decay = [], []
    if isinstance(module, list):
        for entry in module:
            tmp = separate_parameters(entry, skip_list=skip_list)
            decay.extend(tmp[0])
            no_decay.extend(tmp[1])
    elif isinstance(module, torch.nn.Parameter):
        no_decay.append(module)
    else:
        if isinstance(module, torch.nn.Identity):
            return decay, no_decay
        if isinstance(
            module,
            (
                torch.nn.GroupNorm,
                torch.nn.BatchNorm2d,
                torch.nn.LayerNorm,
                torch.nn.BatchNorm1d,
            ),
        ):
            no_decay.extend(list(module.parameters()))
        else:
            for module_name, submodule in module.named_children():
                if module_name in skip_list:
                    no_decay.extend(list(submodule.parameters()))
                else:
                    tmp = separate_parameters(submodule, skip_list=skip_list)
                    decay.extend(tmp[0])
                    no_decay.extend(tmp[1])
            for name, parameter in module.named_parameters():
                if "." in name:
                    continue
                if name in skip_list:
                    no_decay.append(parameter)
                else:
                    decay.append(parameter)
    return decay, no_decay



import socket
from copy import deepcopy
from datetime import datetime
from typing import Optional, Sequence, Tuple

import torch
import torch.nn.functional as F
import numpy as np
import pytorch_lightning as pl
from pytorch_lightning.callbacks.base import Callback
from pytorch_lightning.callbacks import ModelCheckpoint
# from pytorch_lightning.utilities.exceptions import MisconfigurationException

from .cutmix_utils import cutmix_bbox_and_lam, rand_bbox, rand_bbox_minmax
from .snapmix_utils import get_spm


class RandomAugmentationChoiceCallback(Callback):
    """Randomly pick an augmentation callback to use for each batch.

    Also supports no-op warmups and no-op probability.

    Args:
        callbacks (Sequence[Callback]): A sequence of calbacks to choose from.
        p (Sequence[Callback]): A sequence of probabilities for the callbacks.
        no_op_warmup (int, optional): the number of initial steps that should not have any augmentation. Defaults to 0.
        no_op_prob (float, optional): the probability of a step that has no augmentation. Defaults to 0.
    """
    def __init__(
            self, callbacks: Sequence[Callback], p: Sequence[Callback],
            no_op_warmup: int = 0, no_op_prob: float = 0):
        self.p = np.asarray(p) / np.sum(p)
        self.callbacks = callbacks
        self.no_op_warmup = no_op_warmup
        self.step = 0
        self.no_op_prob = no_op_prob
        assert len(p) == len(callbacks)

    def get_callback(self):
        return np.random.choice(self.callbacks, p=self.p)

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self.step += 1
        if self.no_op_warmup >= self.step:
            return
        if self.no_op_prob and np.random.random() < self.no_op_prob:
            return
        self.get_callback().on_train_batch_start(
            trainer, pl_module, batch, batch_idx, dataloader_idx)


class SnapMixCallback(Callback):
    """Callback that perform SnapMix augmentation on the input batch.

    Reference: `Shaoli-Huang/SnapMix <https://github.com/Shaoli-Huang/SnapMix/>`_

    Warning:
        1. **Requires the model to have implemented `extract_features` and `get_fc` methods.**
        2. Can only run in CUDA-enabled environments.
    """

    def __init__(
            self, model, image_size, half: bool = False,
            minmax: Optional[Tuple[float, float]] = None,
            cutmix_bbox: bool = False,  # cutmix style randbox
            alpha: float = 0.4, softmax_target: bool = True):
        self._model = model
        self._half = half
        self.image_size = image_size
        self.alpha = alpha
        self.minmax = minmax
        self.cutmix_bbox = cutmix_bbox
        assert softmax_target, "SnapMix only support softmax_target=True"

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        old_batch = batch
        batch, targets = batch
        target_activation_map = get_spm(
            batch, targets, self._model, self.image_size, self._half
        )
        bs = batch.size(0)
        lamb_1 = np.clip(np.random.beta(self.alpha, self.alpha), 0.05, 0.95)
        lamb_2 = np.clip(np.random.beta(self.alpha, self.alpha), 0.05, 0.95)
        rand_index = torch.randperm(bs).cuda()
        target_activation_map_b = target_activation_map[rand_index, :, :]

        # same_label = target == target_b

        area_1, area_2, cnt = 0, 0, 0
        while area_1 <= 0 or area_2 <= 0:
            if self.minmax:
                bby1_1, bby2_1, bbx1_1, bbx2_1 = rand_bbox_minmax(
                    batch.size(), self.minmax)
                bby1_2, bby2_2, bbx1_2, bbx2_2 = rand_bbox_minmax(
                    batch.size(), self.minmax)
            else:
                bby1_1, bby2_1, bbx1_1, bbx2_1 = rand_bbox(batch.size(), lamb_1)
                bby1_2, bby2_2, bbx1_2, bbx2_2 = rand_bbox(batch.size(), lamb_2)
            if self.cutmix_bbox is True:
                # Use only one random bounding box
                bby1_2, bby2_2, bbx1_2, bbx2_2 = bby1_1, bby2_1, bbx1_1, bbx2_1
            area_1 = (bby2_1-bby1_1) * (bbx2_1-bbx1_1)
            area_2 = (bby2_2-bby1_2) * (bbx2_2-bbx1_2)
            cnt += 1
            # Avoid infinite loops when something goes wrong
            assert cnt < 10, f"{lamb_1}, {lamb_2}"

        cropped = batch[rand_index, :, bby1_2:bby2_2, bbx1_2:bbx2_2].clone()
        if self.cutmix_bbox is False:
            cropped = F.interpolate(
                cropped,
                size=(bby2_1-bby1_1, bbx2_1-bbx1_1),
                mode='bilinear',
                align_corners=True
            )
        batch[:, :, bby1_1:bby2_1, bbx1_1:bbx2_1] = cropped
        lamb_1 = (
            1 - target_activation_map[
                :, bby1_1:bby2_1, bbx1_1:bbx2_1
            ].sum(1).sum(1)
        )
        lamb_2 = target_activation_map_b[
            :, bby1_2:bby2_2, bbx1_2:bbx2_2
        ].sum(1).sum(1)

        # Fall back to Cutmix lambda
        lamb_cutmix = 1 - (
            (bbx2_1 - bbx1_1) * (bby2_1 - bby1_1) /
            (batch.size(2) * batch.size(3))
        )
        lamb_1[torch.isnan(lamb_1)] = lamb_cutmix
        lamb_2[torch.isnan(lamb_2)] = 1 - lamb_cutmix
        # Combine targets
        new_targets = torch.stack([
            targets.float(), targets[rand_index].float(),
            lamb_1.to(targets.device), lamb_2.to(targets.device)
        ], dim=1)
        old_batch[0] = batch
        old_batch[1] = new_targets


class CutMixCallback(Callback):
    """Callback that perform CutMix augmentation on the input batch.

    Assumes the first dimension is batch.

    Reference: `rwightman/pytorch-image-models/ <https://github.com/rwightman/pytorch-image-models/blob/8c9814e3f500e8b37aae86dd4db10aba2c295bd2/timm/data/mixup.py>`_
    """

    def __init__(self, alpha: float = 0.4, softmax_target: bool = False, minmax: Optional[Tuple[float, float]] = None):
        super().__init__()
        self.alpha = alpha
        self.softmax_target = softmax_target
        self.minmax = minmax

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        old_batch = batch
        batch, targets = batch
        batch_flipped = batch.flip(0).clone()
        lambd = np.random.beta(self.alpha, self.alpha, batch.size(0))
        for i in range(batch.shape[0]):
            (yl, yh, xl, xh), lambd_tmp = cutmix_bbox_and_lam(
                batch.shape, lambd[i], ratio_minmax=self.minmax, correct_lam=True)
            lambd[i] = lambd_tmp
            # fill in the cut regions
            batch[i, :, yl:yh, xl:xh] = batch_flipped[i, :, yl:yh, xl:xh]
        # Create the tensor and expand (for target)
        lambd_tensor = batch.new(lambd).view(
            -1, *[1 for _ in range(len(targets.size())-1)]
        ).expand(-1, *targets.shape[1:])
        # Combine targets
        if self.softmax_target:
            new_targets = torch.stack([
                targets.float(), targets.flip(0).float(), lambd_tensor
            ], dim=1)
        else:
            new_targets = (
                targets * lambd_tensor +
                targets.flip(0) * (1-lambd_tensor)
            )
        old_batch[0] = batch
        old_batch[1] = new_targets


class MixUpCallback(Callback):
    """Callback that perform MixUp augmentation on the input batch.

    Assumes the first dimension is batch.

    Works best with pytorch_lightning_spells.losses.MixupSoftmaxLoss

    Reference: `Fast.ai's implementation <https://github.com/fastai/fastai/blob/master/fastai/callbacks/mixup.py>`_
    """

    def __init__(self, alpha: float = 0.4, softmax_target: bool = False):
        super().__init__()
        self.alpha = alpha
        self.softmax_target = softmax_target

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        old_batch = batch
        batch, targets = batch
        batch_flipped = batch.flip(0).clone()
        lambd = np.random.beta(self.alpha, self.alpha, batch.size(0))
        lambd = np.concatenate(
            [lambd[:, np.newaxis], 1-lambd[:, np.newaxis]], axis=1
        ).max(axis=1)
        # Create the tensor and expand (for batch inputs)
        lambd_tensor = batch.new(lambd).view(
            -1, *[1 for _ in range(len(batch.size())-1)]
        ).expand(-1, *batch.shape[1:])
        # Combine input batch
        new_batch = (batch * lambd_tensor +
                     batch_flipped * (1-lambd_tensor))
        # Create the tensor and expand (for target)
        lambd_tensor = batch.new(lambd).view(
            -1, *[1 for _ in range(len(targets.size())-1)]
        ).expand(-1, *targets.shape[1:])
        # Combine targets
        if self.softmax_target:
            new_targets = torch.stack([
                targets.float(), targets.flip(0).float(), lambd_tensor
            ], dim=1)
        else:
            new_targets = (
                targets * lambd_tensor +
                targets.flip(0) * (1-lambd_tensor)
            )
        old_batch[0] = new_batch
        old_batch[1] = new_targets


class TelegramCallback(Callback):
    """A Telegram notification callback

    Reference: `huggingface/knockknock <https://github.com/huggingface/knockknock>`_
    """
    DATE_FORMAT = "%Y-%m-%d %H:%M:%d"

    def __init__(self, token: str, chat_id: int, name: str, report_evals: bool = False):
        try:
            import telegram
        except ImportError:
            raise ImportError(
                "Please install 'python-telegram-bot' before using TelegramCallback.")
        self._token = token
        self.telegram_bot = telegram.Bot(token=self._token)
        self.host_name = socket.gethostname()
        self.report_evals = report_evals
        self.chat_id = chat_id
        self.name = name
        self.start_time = datetime.now()

    def send_message(self, text):
        try:
            self.telegram_bot.send_message(chat_id=self.chat_id, text=text)
        except telegram.error.TimeOut:
            # Ignore timeouts and continue training
            pass

    def on_train_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        self.start_time = datetime.now()
        contents = [
            f'{self.name} has started training 🎬',
            'Machine name: %s' % self.host_name,
            'Starting date: %s' % self.start_time.strftime(
                TelegramCallback.DATE_FORMAT)
        ]
        text = '\n'.join(contents)
        self.send_message(text=text)

    def on_train_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        end_time = datetime.now()
        elapsed_time = end_time - self.start_time
        contents = [
            f'{self.name} has finished training 🎉',
            'Machine name: %s' % self.host_name,
            'Starting date: %s' % self.start_time.strftime(
                TelegramCallback.DATE_FORMAT),
            'End date: %s' % end_time.strftime(
                TelegramCallback.DATE_FORMAT),
            'Training duration: %s' % str(elapsed_time)
        ]
        text = '\n'.join(contents)
        self.send_message(text=text)

    def on_validation_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        metrics, meta = self._collect_metrics(trainer)
        if self.report_evals is False:
            return
        contents = [
            f"Metrics from {self.name} at step {meta['step']} (epoch {meta['epoch']}):"
        ]
        contents += [
            f"{metric_name}: {metric_value:.6f}"
            for metric_name, metric_value in metrics.items()
            if metric_name != "epoch"
        ]
        text = '\n'.join(contents)
        self.send_message(text=text)

    def _collect_metrics(self, trainer):
        ckpt_name_metrics = deepcopy(trainer.logger_connector.logged_metrics)
        # ckpt_name_metrics.update(trainer.logger_connector.callback_metrics)
        # ckpt_name_metrics.update(trainer.logger_connector.progress_bar_metrics)
        meta = {"step": trainer.global_step, "epoch": trainer.current_epoch}
        return ckpt_name_metrics, meta


class LookaheadCallback(Callback):
    """Switch to the slow weights before evaluation and switch back after.
    """

    def on_validation_start(self, trainer, pl_module):
        optimizer = trainer.optimizers(use_pl_optimizer=False)
        if hasattr(optimizer, "_backup_and_load_cache"):
            print("load slow parameters")
            optimizer._backup_and_load_cache()

    def on_validation_end(self, trainer, pl_module):
        optimizer = trainer.optimizers(use_pl_optimizer=False)
        if hasattr(optimizer, "_clear_and_load_backup"):
            print("load fast parameters")
            optimizer._clear_and_load_backup()


class LookaheadModelCheckpoint(ModelCheckpoint):
    """Combines LookaheadCallback and ModelCheckpoint
    """

    def on_validation_start(self, trainer, pl_module):
        for optimizer in trainer.optimizers:
            if hasattr(optimizer, "_backup_and_load_cache"):
                print("load slow parameters")
                optimizer._backup_and_load_cache()

    def on_validation_end(self, trainer, pl_module):
        self.save_checkpoint(trainer, pl_module)
        for optimizer in trainer.optimizers:
            if hasattr(optimizer, "_clear_and_load_backup"):
                print("load fast parameters")
                optimizer._clear_and_load_backup()



import pytorch_lightning as pl

from . import callbacks
from . import loggers
from . import losses
from . import optimizers
from . import utils
from . import lr_schedulers
from . import metrics
from . import samplers
from .version import (
    __version__,
    __docs__,
    __author__,
    __author_email__,
    __license__
)


class BaseModule(pl.LightningModule):
    """A boilerplate module with some sensible defaults.

    It logs the exponentially smoothed training losses, and the validation metrics.

    You need to implement the `training_step()` and `validation_step()` methods.
    Please refer to the `training_step_end()` and `validation_step_end()` methods for the expected output format.

    Args:
        ema_alpha (float, optional): the weight of the new training loss for the EMA aggregator. Defaults to 0.02.

    Example:

        >>> class TestModule(BaseModule):
        ...     def __init__(self):
        ...         super().__init__(ema_alpha=0.02)
        ...
        ...     def training_step(self, batch, batch_idx):
        ...         return {
        ...             "loss": torch.tensor(1),
        ...             "log": batch_idx % self.trainer.accumulate_grad_batches == 0
        ...         }
        ...
        ...     def validation_step(self, batch, batch_idx):
        ...         return {
        ...             'loss': torch.tensor(1),
        ...             'preds': torch.ones_like(batch[1]),
        ...             'target': batch[1]
        ...         }
        ...
        >>> module = TestModule()
    """

    def __init__(self, ema_alpha: float = 0.02):
        super().__init__()
        self.train_loss_tracker = utils.EMATracker(ema_alpha)

    def get_progress_bar_dict(self):
        """Removes `v_num` from the progress bar.
        """
        # don't show the experiment version number
        items = super().get_progress_bar_dict()
        items.pop("v_num", None)
        return items

    def _should_log(self, flag):
        """Determines if the loss of this training step should be logged.

        Args:
            flag (Union[bool, List[bool]]): if this is a loggable step.

        Returns:
            bool: True if this step should be logged.
        """
        if (self.trainer.global_step + 1) % self.trainer.log_every_n_steps == 0:
            if isinstance(flag, list):
                # for distributed training scenarios
                return flag[0]
            return flag
        return False

    def training_step_end(self, outputs):
        """This method logs the training loss for you.

        It follows the `log_every_n_steps` attribute of the associated Trainer.

        The output from `.validation_step()` method must contains these two entries:

        1. loss: the training loss.
        2. log: a boolean value indicating if this is a loggable step.


        A loggable step is a step that involves an optimizer step. The opposite is a step that only updates the gradients but not the parameters(e.g., in gradient accumulation).

        Args:
            outputs (Dict): the output from `.training_step()` method.
        """
        loss = outputs["loss"].mean()
        self.train_loss_tracker.update(loss.detach())
        if self._should_log(outputs["log"]):
            self.logger.log_metrics({
                "train_loss": self.train_loss_tracker.value
            }, step=self.global_step)
        return loss

    def validation_step_end(self, outputs):
        """This method logs the validation loss and metrics for you.

        The output from `.validation_step()` method must contains these three entries:

            1. loss: the validation loss.
            2. pred: the predicted labels or values.
            3. target: the ground truth lables or values.

        Args:
            outputs (Dict): the output from `.validation_step()` method.
        """
        self.log('val_loss', outputs['loss'].mean())
        for name, metric in self.metrics:
            metric(
                outputs['pred'].view(-1).cpu(),
                outputs['target'].view(-1).cpu()
            )
            self.log("val_" + name, metric)

    def test_step(self, batch, batch_idx):
        """Simply defer to `.validation_step()` method."""
        return self.validation_step(batch, batch_idx)

    def test_step_end(self, outputs):
        """Basically the same as `.validation_step_ends()` method, but with a different prefix.
        """
        # TODO: refactor to be able to simply defer to `.validation_step_ends()`?
        self.log('test_loss', outputs['loss'].mean())
        for name, metric in self.metrics:
            metric(
                outputs['pred'].view(-1).cpu(),
                outputs['target'].view(-1).cpu()
            )
            self.log("test_" + name, metric)



"""SnapMix Utility Functions

Reference: `Shaoli-Huang/SnapMix <https://github.com/Shaoli-Huang/SnapMix/>`_
"""
import torch
import torch.nn.functional as F


def get_spm(input_tensor, target, model, image_size, half: bool = False):
    bs = input_tensor.size(0)
    with torch.no_grad():
        fc = model.get_fc()
        if half:
            input_tensor = input_tensor.half()
        fms = model.extract_features(input_tensor.to(fc.weight.device))
        weight, bias = fc.weight, fc.bias
        weight = weight.view(weight.size(0), weight.size(1), 1, 1)

        activations = F.conv2d(fms, weight, bias=bias)

        tmp = []
        for i in range(bs):
            target_activation = activations[i, target[i]]
            tmp.append(target_activation)

        target_activation_map = torch.stack(tmp)
        target_activation_map = target_activation_map.unsqueeze(1)
        target_activation_map = F.interpolate(
            target_activation_map, image_size,
            mode='bilinear', align_corners=False)

        target_activation_map = target_activation_map.squeeze(1)

        for i in range(bs):
            target_activation_map[i] -= target_activation_map[i].min()
            target_activation_map[i] /= target_activation_map[i].sum()

    return target_activation_map



from typing import Tuple, Optional, Union, List

import numpy as np


def rand_bbox(img_shape:Tuple, lam: float, margin: float=0., count: Optional[int]=None):
    """ Standard CutMix bounding-box

    Generates a random square bbox based on lambda value. This impl includes
    support for enforcing a border margin as percent of bbox dimensions.

    Args:
        img_shape: Image shape as tuple
        lam: Cutmix lambda value
        margin: Percentage of bbox dimension to enforce as margin (reduce amount of box outside image)
        count: Number of bbox to generate
    """
    ratio = np.sqrt(1 - lam)
    img_h, img_w = img_shape[-2:]
    cut_h, cut_w = int(img_h * ratio), int(img_w * ratio)
    margin_y, margin_x = int(margin * cut_h), int(margin * cut_w)
    cy = np.random.randint(0 + margin_y, img_h - margin_y, size=count)
    cx = np.random.randint(0 + margin_x, img_w - margin_x, size=count)
    yl = np.clip(cy - cut_h // 2, 0, img_h)
    yh = np.clip(cy + cut_h // 2, 0, img_h)
    xl = np.clip(cx - cut_w // 2, 0, img_w)
    xh = np.clip(cx + cut_w // 2, 0, img_w)
    return yl, yh, xl, xh


def rand_bbox_minmax(img_shape: Tuple, minmax: Union[Tuple, List], count: Optional[int]=None):
    """ Min-Max CutMix bounding-box

    Inspired by Darknet cutmix impl, generates a random rectangular bbox
    based on min/max percent values applied to each dimension of the input image.
    Typical defaults for minmax are usually in the  .2-.3 for min and .8-.9 range for max.

    Args:
        img_shape: Image shape as tuple
        minmax: Min and max bbox ratios (as percent of image size)
        count: Number of bbox to generate
    """
    assert len(minmax) == 2
    img_h, img_w = img_shape[-2:]
    cut_h = np.random.randint(
        int(img_h * minmax[0]), int(img_h * minmax[1]), size=count)
    cut_w = np.random.randint(
        int(img_w * minmax[0]), int(img_w * minmax[1]), size=count)
    yl = np.random.randint(0, img_h - cut_h, size=count)
    xl = np.random.randint(0, img_w - cut_w, size=count)
    yu = yl + cut_h
    xu = xl + cut_w
    return yl, yu, xl, xu


def cutmix_bbox_and_lam(img_shape: Tuple, lam, ratio_minmax=None, correct_lam=True, count: Optional[int]=None):
    """ Generate bbox and apply lambda correction.
    """
    if ratio_minmax is not None:
        yl, yu, xl, xu = rand_bbox_minmax(img_shape, ratio_minmax, count=count)
    else:
        yl, yu, xl, xu = rand_bbox(img_shape, lam, count=count)
    if correct_lam or ratio_minmax is not None:
        bbox_area = (yu - yl) * (xu - xl)
        lam = 1. - bbox_area / float(img_shape[-2] * img_shape[-1])
    return (yl, yu, xl, xu), lam



import numpy as np
from torch.utils.data.sampler import Sampler


class SortishSampler(Sampler):
    """Go through the text data by order of length with a bit of randomness.


    Returns an iterator that traverses the the data in randomly ordered batches that are approximately the same size.

    The data is first randomly shuffled and then put into a number of chunks. The data in each chunk is then sorted and
    sliced to get batches that are approximately the same size.

    The max key size batch is always returned in the first call because of pytorch cuda memory allocation sequencing.

    Without that max key returned first multiple buffers may be allocated when the first created isn't large enough
    to hold the next in the sequence.

    Taken from `Fast.ai <https://github.com/fastai/fastai1/blob/bcef12e95405655481bb309761f8c552b51b2bd2/fastai/text/data.py#L107>`_.

    Args:
        data_source (Iterable): The data you want to sample from.
        key (Callable): A function to get keys to sort. Input: the index number of the entry in data_source.
        bs (int): the batch size for the data loader
        chunk_size (int, optional): the number of batches one chunk contains. Defaults to 100.

    Example:
        >>> data_source = [[0], [0, 1]] * 100
        >>> sampler = SortishSampler(data_source, key=lambda idx: len(data_source[idx]), bs=2, chunk_size=2)
        >>> len(list(sampler))
        200
        >>> len(data_source[next(iter(sampler))]) # the largest/longest batch always goes first
        2
    """

    def __init__(self, data_source, key, bs, chunk_size=100):
        self.data_source, self.key, self.bs = data_source, key, bs
        self.chunk_size = chunk_size

    def __len__(self): return len(self.data_source)

    def __iter__(self):
        while True:
            idxs = np.random.permutation(len(self.data_source))
            sz = self.bs * self.chunk_size
            ck_idx = [idxs[i:i+sz] for i in range(0, len(idxs), sz)]
            # Sort inside the chunk (from longest to shortest)
            sort_idx = np.concatenate(
                [sorted(s, key=self.key, reverse=True) for s in ck_idx])
            sz = self.bs
            # Get batches
            ck_idx = [sort_idx[i:i+sz] for i in range(0, len(sort_idx), sz)]
            # find the batch with the largest key,
            max_ck = np.argmax([self.key(ck[0]) for ck in ck_idx])
            # then make sure it goes first.
            if len(ck_idx[max_ck]) != self.bs:
                # if not a full batch, reshuffle again
                continue
            # swap the batch
            ck_idx[0], ck_idx[max_ck] = ck_idx[max_ck], ck_idx[0]
            # Shuffle the list of batches
            sort_idx = np.concatenate(np.random.permutation([
                np.random.permutation(chunk.reshape(self.bs, -1)).reshape(-1)
                for chunk in ck_idx[1:-1]
            ]))
            sort_idx = np.concatenate((ck_idx[0], sort_idx, ck_idx[-1]))
            break
        return iter(sort_idx)


class SortSampler(Sampler):
    """Go through the text data by order of length (longest to shortest).

    Taken from `Fast.ai library <https://github.com/fastai/fastai1/blob/bcef12e95405655481bb309761f8c552b51b2bd2/fastai/text/data.py#L99>`_.

    Args:
        data_source (Iterable): The data you want to sample from.
        key (Callable): A function to get keys to sort. Input: the index number of the entry in data_source.

    Example:
        >>> data_source = [[0], [0, 1], [0, 1, 2, 3]]
        >>> sampler = SortSampler(data_source, key=lambda idx: len(data_source[idx]))
        >>> len(list(sampler))
        3
        >>> next(iter(sampler)) # the longest entry is the third one.
        2
    """

    def __init__(self, data_source, key):
        self.data_source, self.key = data_source, key

    def __len__(self):
        return len(self.data_source)

    def __iter__(self):
        return iter(sorted(
            range(len(self.data_source)),
            key=self.key, reverse=True
        ))



#!/usr/bin/env python
import os

from setuptools import find_packages, setup

try:
    from pytorch_lightning_spells import version as pls
except (ImportError, ModuleNotFoundError):
    # alternative https://stackoverflow.com/a/67692/4521646
    import sys
    sys.path.append("pytorch_lightning_spells")
    import version as pls

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="pytorch-lightning-spells",
    version=pls.__version__,
    description=pls.__docs__,
    author=pls.__author__,
    author_email=pls.__author_email__,
    url='https://github.com/veritable-tech/pytorch-lightning-spells',
    download_url='https://github.com/veritable-tech/pytorch-lightning-spells',
    license=pls.__license__,
    packages=find_packages(exclude=['tests', 'tests/*', 'benchmarks']),
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['deep learning', 'pytorch', 'AI'],
    python_requires='>=3.6',
    setup_requires=[],
    install_requires=["pytorch_lightning>=1.1.3, <=1.4.0", "scipy", "scikit-learn"],
    extras_require={},
    classifiers=[
        'Environment :: Console',
        'Natural Language :: English',
        # How mature is this project? Common values are
        #   3 - Alpha, 4 - Beta, 5 - Production/Stable
        'Development Status :: 4 - Beta',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Scientific/Engineering :: Information Analysis',
        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



import weakref
from functools import wraps
from typing import Sequence, Union

import torch
import numpy as np
from torch.optim.lr_scheduler import _LRScheduler, CosineAnnealingLR
from torch.optim import Optimizer

__all__ = [
    "BaseLRScheduler", "LinearLR",
    "ExponentialLR", "MultiStageScheduler",
    "CosineAnnealingScheduler"
]


class BaseLRScheduler(_LRScheduler):
    def switch_optimizer(self, optimizer):
        self.optimizer = optimizer
        self.optimizer._step_count = self._step_count

    def clear_optimizer(self):
        self.optimizer = None


class CosineAnnealingScheduler(CosineAnnealingLR, BaseLRScheduler):
    pass


class LinearLR(BaseLRScheduler):
    """Linearly increases or decrease the learning rate between two boundaries over a number of
    iterations.
    """

    def __init__(self, optimizer: torch.optim.Optimizer, min_lr_ratio: float, total_epochs: float, upward: bool=True, last_epoch: int=-1):
        """Initialize a scheduler.

        Args:
            optimizer (Union[torch.optim.Optimizer, apex.fp16_utils.fp16_optimizer.FP16_Optimizer]):
            min_lr_ratio:  min_lr_ratio * base_lr will be the starting learning rate.
            total_epochs: the total number of "steps" in this run.
            upward: whether the learning rate goes up or down. Defaults to True.
            last_epoch: the index of last epoch. Defaults to -1.
        """
        assert min_lr_ratio < 1
        self.upward = upward
        self.min_lr_ratio = min_lr_ratio
        self.total_epochs = total_epochs - 1  # starts at zero
        super(LinearLR, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        current_epoch = self.last_epoch
        if self.upward:
            progress = 1 - current_epoch / self.total_epochs  # 1 to 0
        else:
            progress = current_epoch / self.total_epochs  # 1 to 0
        # safety measure
        progress = max(min(progress, 1.), 0.)
        return [
            base_lr - progress * (base_lr - self.min_lr_ratio * base_lr)
            for base_lr in self.base_lrs
        ]


class ExponentialLR(BaseLRScheduler):
    """Exponentially increases the learning rate between two boundaries over
    a number of iterations.

    Mainly used by LR finders.
    """

    def __init__(self, optimizer, min_lr_ratio, total_epochs, last_epoch=-1):
        """Initialize a scheduler.

        Parameters
        ----------
        optimizer : Union[torch.optim.Optimizer, apex.fp16_utils.fp16_optimizer.FP16_Optimizer]
        min_lr_ratio : float
            min_lr_ratio * base_lr will be the starting learning rate.
        total_epochs : int
            the total number of "steps" in this run.
        last_epoch : int, optional
            the index of last epoch, by default -1.
        """
        assert min_lr_ratio < 1
        self.min_lr_ratio = min_lr_ratio
        self.total_epochs = total_epochs - 1  # start from zero
        super(ExponentialLR, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        current_epoch = self.last_epoch + 1
        progress = 1 - current_epoch / self.total_epochs  # 1 to 0
        return [base_lr * (self.min_lr_ratio) ** progress for base_lr in self.base_lrs]


class MultiStageScheduler(_LRScheduler):
    def __init__(self, schedulers: Sequence, start_at_epochs: Sequence[int], last_epoch: int = -1):
        assert len(schedulers) == len(start_at_epochs)
        schedulers, start_at_epochs = (
            np.array(schedulers), np.array(start_at_epochs))
        # sort starting epochs in descending order
        idx = np.flip(np.argsort(start_at_epochs))
        self.schedulers = schedulers[idx]
        self.start_at_epochs = start_at_epochs[idx]
        self.last_epoch = last_epoch
        # Explicitly run step(). Otherwise the initial LR will be initialized by the last sub-scheduler
        self.step(0)
        self.optimizer = self.schedulers[0].optimizer

    def step(self, epoch=None):
        if epoch is None:
            self.last_epoch = self.last_epoch + 1
        else:
            self.last_epoch = epoch - 1
        for scheduler, starting_epoch in zip(self.schedulers, self.start_at_epochs):
            if self.last_epoch + 1 >= starting_epoch:
                scheduler.last_epoch = self.last_epoch - starting_epoch
                return scheduler.step()

    def switch_optimizer(self, optimizer):
        for scheduler in self.schedulers:
            scheduler.optimizer = optimizer
            scheduler.optimizer._step_count = scheduler._step_count
        self.optimizer = self.schedulers[0].optimizer

    def clear_optimizer(self):
        for scheduler in self.schedulers:
            scheduler.optimizer = None
        self.optimizer = None



from typing import Dict, Optional
from pytorch_lightning.loggers.base import LightningLoggerBase
from pytorch_lightning.utilities.distributed import rank_zero_only


class ScreenLogger(LightningLoggerBase):
    """A logger that prints metrics to the screen.

    Suitable in situation where you want to check the training progress directly in the console.
    """

    def __init__(self):
        super().__init__(self)

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        if any((key.startswith("val_") for key in metrics.keys())):
            print("")
        for key, val in metrics.items():
            if key.startswith("val_"):
                print(key, "%.4f" % val)

    @rank_zero_only
    def log_hyperparams(self, params):
        pass

    @property
    def experiment(self):
        return None

    @property
    def name(self):
        return "screen_logger"

    @property
    def version(self) -> int:
        return 0
